@@ -71,6 +71,7 @@ a :py:class:`Model`.
 # Author: Nathan Trouvain at 01/06/2021 <nathan.trouvain@inria.fr>
 # Licence: MIT License
 # Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
+from collections import defaultdict
 from contextlib import ExitStack, contextmanager
 from functools import partial
 from typing import Dict, List, Optional, Sequence, Tuple
@@ -83,21 +84,24 @@ from .type import MappedData
 from .utils import progress, verbosity
 from .utils.graphflow import (
     DataDispatcher,
+    dispatch,
     find_entries_and_exits,
     get_offline_subgraphs,
     topological_sort,
 )
 from .utils.model_utils import (
-    _allocate_returned_states,
-    _build_forward_sumodels,
-    _dist_states_to_next_subgraph,
-    to_ragged_seq_set,
+    allocate_returned_states,
+    build_forward_sumodels,
+    dist_states_to_next_subgraph,
+    fold_mapping,
+    to_data_mapping,
 )
 from .utils.validation import is_mapping
 
 
-def _run_and_partial_fit(
-    model,
+def run_and_partial_fit(
+    submodel,
+    complete_model,
     offlines,
     relations,
     x_seq,
@@ -110,38 +114,77 @@ def _run_and_partial_fit(
 ):
     """Run a submodel and call its partial fit function."""
 
-    if not model.is_empty:
-        x_seq = {n: x for n, x in x_seq.items() if n in model.node_names}
+    if not submodel.is_empty:
+        x_seq = {n: x for n, x in x_seq.items() if n in submodel.node_names}
 
-        if force_teachers:
-            y_seq = {n: y for n, y in y_seq.items() if n in [o.name for o in offlines]}
+        if force_teachers and y_seq is not None:
+            y_seq = {
+                n: y
+                for n, y in y_seq.items()
+                if n in [o.name for o in complete_model.nodes if o.is_trained_offline]
+            }
         else:
             y_seq = None
 
-        model._is_initialized = True
-        states = model.run(
-            x_seq, y_seq, return_states=return_states, stateful=stateful, reset=reset
+        submodel._is_initialized = True
+        states = run_submodel(
+            complete_model,
+            submodel,
+            x_seq,
+            y_seq,
+            return_states=return_states,
+            stateful=stateful,
+            reset=reset,
         )
 
-        dist_states = _dist_states_to_next_subgraph(states, relations)
+        dist_states = dist_states_to_next_subgraph(states, relations)
     else:
         dist_states = x_seq
 
-    for node in offlines:
-        node.partial_fit(dist_states[node.name], y_seq[node.name], warmup=warmup)
-
-        return dist_states
-
-
-def _filter_teacher_nodes(Y):
-    if is_mapping(Y):
-        sequences = {n: v for n, v in Y.items() if not isinstance(v, _Node)}
-        teachers = {n: v for n, v in Y.items() if isinstance(v, _Node)}
-        return sequences, teachers
-    elif isinstance(Y, _Node):
-        return None, Y
+    if y_seq is not None:
+        for node in offlines:
+            node.partial_fit(
+                dist_states.get(node.name), y_seq.get(node.name), warmup=warmup
+            )
     else:
-        return Y, None
+        for node in offlines:
+            node.partial_fit(dist_states.get(node.name), warmup=warmup)
+
+    return dist_states
+
+
+def run_submodel(
+    model,
+    submodel,
+    X: MappedData,
+    forced_feedbacks: Dict[str, np.ndarray] = None,
+    from_state: Dict[str, np.ndarray] = None,
+    stateful=True,
+    reset=False,
+    shift_fb=True,
+    return_states: Sequence[str] = None,
+) -> MappedData:
+
+    X_, forced_feedbacks_ = to_data_mapping(submodel, X, forced_feedbacks)
+
+    submodel._initialize_on_sequence(X_[0], forced_feedbacks_[0])
+
+    states = []
+    for X_seq, fb_seq in zip(X_, forced_feedbacks_):
+        with model.with_state(reset=reset, stateful=stateful):
+            states_seq = model._run(
+                X_seq,
+                fb_seq,
+                from_state=from_state,
+                stateful=stateful,
+                shift_fb=shift_fb,
+                return_states=return_states,
+                submodel=submodel,
+            )
+
+            states.append(states_seq)
+
+    return fold_mapping(submodel, states, return_states)
 
 
 def forward(model: "Model", x: MappedData) -> List[np.ndarray]:
@@ -266,6 +309,8 @@ class Model(_Node):
 
         self._name = self._get_name(name)
 
+        nodes, edges = self._concat_multi_inputs(nodes, edges)
+
         self._edges = edges
 
         # always maintain nodes in topological order
@@ -273,7 +318,8 @@ class Model(_Node):
             self._inputs, self._outputs = find_entries_and_exits(nodes, edges)
             self._nodes = topological_sort(nodes, edges, self._inputs)
         else:
-            self._inputs = self._outputs = list()
+            self._inputs = list()
+            self._outputs = list()
             self._nodes = nodes
 
         self._is_initialized = False
@@ -295,11 +341,17 @@ class Model(_Node):
     def __getitem__(self, item):
         return self.get_node(item)
 
-    def __iand__(self, other):
+    def __iand__(self, other) -> "Model":
         from .ops import merge
 
         return merge(self, other, inplace=True)
 
+    @staticmethod
+    def _concat_multi_inputs(nodes, edges):
+        from .ops import concat_multi_inputs
+
+        return concat_multi_inputs(nodes, edges)
+
     def _check_if_only_online(self):
         if any([n.is_trained_offline and not n.fitted for n in self.nodes]):
             raise RuntimeError(
@@ -338,28 +390,73 @@ class Model(_Node):
 
             self.initialize(x_init, y_init)
 
-    def _call(self, x=None, return_states=None, *args, **kwargs):
+    def _call(self, x=None, return_states=None, submodel=None, *args, **kwargs):
 
-        self._forward(self, x)
+        if submodel is None:
+            submodel = self
+
+        self._forward(submodel, x)
 
         state = {}
         if return_states == "all":
-            for node in self.nodes:
+            for node in submodel.nodes:
                 state[node.name] = node.state()
 
         elif hasattr(return_states, "__iter__"):
             for name in return_states:
-                state[name] = self[name].state()
+                state[name] = submodel[name].state()
 
         else:
-            if len(self.output_nodes) > 1:
-                for out_node in self.output_nodes:
+            if len(submodel.output_nodes) > 1:
+                for out_node in submodel.output_nodes:
                     state[out_node.name] = out_node.state()
             else:
-                state = self.output_nodes[0].state()
+                state = submodel.output_nodes[0].state()
 
         return state
 
+    def _run(
+        self,
+        X,
+        feedback,
+        from_state=None,
+        stateful=True,
+        shift_fb=True,
+        return_states=None,
+        submodel=None,
+    ):
+        if submodel is None:
+            submodel = self
+
+        states = allocate_returned_states(submodel, X, return_states)
+
+        seq = progress(
+            dispatch(X, feedback, shift_fb=shift_fb),
+            f"Running {self.name}",
+            total=len(X),
+        )
+
+        with self.with_state(from_state, stateful=stateful):
+            # load proxys after state update to make it have an
+            # impact on feedback
+            self._load_proxys(keep=True)
+            for i, (x, forced_fb, _) in enumerate(seq):
+
+                with self.with_feedback(forced_fb):
+                    state = submodel._call(x, return_states=return_states)
+
+                if is_mapping(state):
+                    for name, value in state.items():
+                        states[name][i, :] = value
+                else:
+                    states[submodel.output_nodes[0].name][i, :] = state
+
+                self._load_proxys()
+
+        self._clean_proxys()
+
+        return states
+
     def _unregister_teachers(self):
         """Remove teacher nodes refs from student nodes."""
         for node in self.trainable_nodes:
@@ -385,8 +482,13 @@ class Model(_Node):
         Model
             The updated Model.
         """
-        self._nodes = list(set(new_nodes) | set(self.nodes))
-        self._edges = list(set(new_edges) | set(self.edges))
+        nodes = list(set(new_nodes) | set(self.nodes))
+        edges = list(set(new_edges) | set(self.edges))
+
+        nodes, edges = self._concat_multi_inputs(nodes, edges)
+
+        self._nodes = nodes
+        self._edges = edges
 
         self._params = {n.name: n.params for n in self._nodes}
         self._hypers = {n.name: n.hypers for n in self._nodes}
@@ -619,14 +721,21 @@ class Model(_Node):
             yield self
             return
 
-        with ExitStack() as stack:
-            if feedback is not None:
+        elif feedback is not None:
+            with ExitStack() as stack:
                 for node in self.nodes:
                     value = feedback.get(node.name)
+                    # maybe node does not send feedback but receives it?
+                    if value is None and node.has_feedback:
+                        value = feedback.get(node._feedback.name)
                     stack.enter_context(
                         node.with_feedback(value, stateful=stateful, reset=reset)
                     )
+
+                yield self
+        else:
             yield self
+            return
 
     def reset(self, to_state: Dict[str, np.ndarray] = None):
         """Reset the last state saved to zero for all Nodes in
@@ -773,7 +882,7 @@ class Model(_Node):
             A sequence of input data.
             If dict, then pairs of keys and values, where keys are Model input
             nodes names and values are sequence of input data.
-        forced_feedback: dict of array-like of shape ([n_feedbacks], timesteps, feedabck_dim)
+        forced_feedbacks: dict of array-like of shape ([n_feedbacks], timesteps, feedabck_dim)
             Pairs of keys and values, where keys are Model nodes names and
             value are sequences of feedback vectors to force into the nodes.
         from_state : dict of arrays of shape ([nodes], 1, nodes.output_dim)
@@ -798,45 +907,25 @@ class Model(_Node):
             where keys are output nodes names and values
             are corresponding sequences of output vectors.
         """
-        X_, forced_feedbacks_ = check_xy(
-            self, X, forced_feedbacks, allow_n_sequences=False
-        )
+        X_, forced_feedbacks_ = to_data_mapping(self, X, forced_feedbacks)
 
-        self._initialize_on_sequence(X_, forced_feedbacks_)
+        self._initialize_on_sequence(X_[0], forced_feedbacks_[0])
 
-        states = _allocate_returned_states(self, X_, return_states)
+        states = []
+        for X_seq, fb_seq in zip(X_, forced_feedbacks_):
+            with self.with_state(reset=reset, stateful=stateful):
+                states_seq = self._run(
+                    X_seq,
+                    fb_seq,
+                    from_state=from_state,
+                    stateful=stateful,
+                    shift_fb=shift_fb,
+                    return_states=return_states,
+                )
 
-        seq = progress(
-            self._dispatcher.dispatch(X_, forced_feedbacks_, shift_fb=shift_fb),
-            f"Running {self.name}",
-            total=len(X_),
-        )
+                states.append(states_seq)
 
-        with self.with_state(from_state, stateful=stateful, reset=reset):
-            # load proxys after state update to make it have an
-            # impact on feedback
-            self._load_proxys(keep=True)
-            for i, (x, forced_fb, _) in enumerate(seq):
-
-                with self.with_feedback(forced_fb):
-                    state = self._call(x, return_states=return_states)
-
-                if is_mapping(state):
-                    for name, value in state.items():
-                        states[name][i, :] = value
-                else:
-                    states[self.output_nodes[0].name][i, :] = state
-
-                self._load_proxys()
-
-        self._clean_proxys()
-
-        # dicts are only relevant if model has several outputs (len > 1) or
-        # if we want to return states from hidden nodes
-        if len(states) == 1 and return_states is None:
-            return states[self.output_nodes[0].name]
-
-        return states
+        return fold_mapping(self, states, return_states)
 
     def train(
         self,
@@ -871,11 +960,6 @@ class Model(_Node):
             the real state of these nodes will be sent to the feedback
             receivers
             during training.
-        call : bool, default to True
-            It True, call the Model and update its Nodes states before
-            applying the
-            learning rule. Otherwise, use the train method
-            on the nodes current states.
         learn_every : int, default to 1
             Time interval at which training must occur, when dealing with a
             sequence of input data. By default, the training method is called
@@ -887,6 +971,8 @@ class Model(_Node):
             If True, Node state will be updated by this operation.
         reset : bool, default to False
             If True, Nodes states will be reset to zero before this operation.
+        return_states: list of str, optional
+            Names of Nodes from which to return states as output.
 
         Returns
         -------
@@ -905,9 +991,9 @@ class Model(_Node):
 
         self._initialize_on_sequence(X_, Y_)
 
-        states = _allocate_returned_states(self, X_, return_states)
+        states = allocate_returned_states(self, X_, return_states)
 
-        dispatch = self._dispatcher.dispatch(
+        dispatched_data = dispatch(
             X_, Y_, return_targets=True, force_teachers=force_teachers
         )
 
@@ -915,7 +1001,7 @@ class Model(_Node):
             # load proxys after state update to make it have an
             # impact on feedback
             self._load_proxys(keep=True)
-            for i, (x, forced_feedback, y) in enumerate(dispatch):
+            for i, (x, forced_feedback, y) in enumerate(dispatched_data):
 
                 if not force_teachers:
                     forced_feedback = None
@@ -999,10 +1085,7 @@ class Model(_Node):
                 "no offline nodes found in model."
             )
 
-        X, Y = to_ragged_seq_set(X), to_ragged_seq_set(Y)
-        data = list(self._dispatcher.dispatch(X, Y, shift_fb=False))
-        X = [datum[0] for datum in data]
-        Y = [datum[1] for datum in data]
+        X, Y = to_data_mapping(self, X, Y)
 
         self._initialize_on_sequence(X[0], Y[0])
         self.initialize_buffers()
@@ -1014,7 +1097,7 @@ class Model(_Node):
 
         with self.with_state(from_state, reset=reset, stateful=stateful):
             for i, ((nodes, edges), relations) in enumerate(subgraphs):
-                submodel, offlines = _build_forward_sumodels(nodes, edges, trained)
+                submodel, offlines = build_forward_sumodels(nodes, edges, trained)
 
                 if next_X is not None:
                     for j in range(len(X)):
@@ -1029,9 +1112,10 @@ class Model(_Node):
                 seq = progress(X, f"Running {self.name}")
 
                 _partial_fit_fn = partial(
-                    _run_and_partial_fit,
+                    run_and_partial_fit,
                     force_teachers=force_teachers,
-                    model=submodel,
+                    complete_model=self,
+                    submodel=submodel,
                     warmup=warmup,
                     reset=reset,
                     offlines=offlines,
