@@ -17,7 +17,12 @@ Metrics and observables for Reservoir Computing:
 # Author: Nathan Trouvain at 01/06/2021 <nathan.trouvain@inria.fr>
 # Licence: MIT License
 # Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
-from typing import Literal
+import sys
+
+if sys.version_info < (3, 8):
+    from typing_extensions import Literal
+else:
+    from typing import Literal
 
 import numpy as np
 from scipy import linalg
@@ -27,6 +32,19 @@ from scipy.sparse.linalg import eigs
 from .type import Weights
 
 
+def _check_arrays(y_true, y_pred):
+    y_true_array = np.asarray(y_true)
+    y_pred_array = np.asarray(y_pred)
+
+    if not y_true_array.shape == y_pred_array.shape:
+        raise ValueError(
+            f"Shape mismatch between y_true and y_pred: "
+            "{y_true_array.shape} != {y_pred_array.shape}"
+        )
+
+    return y_true_array, y_pred_array
+
+
 def spectral_radius(W: Weights, maxiter: int = None) -> float:
     """Compute the spectral radius of a matrix `W`.
 
@@ -93,7 +111,8 @@ def mse(y_true: np.ndarray, y_pred: np.ndarray) -> float:
     float
         Mean squared error.
     """
-    return float(np.mean((np.asanyarray(y_true) - np.asanyarray(y_pred)) ** 2))
+    y_true_array, y_pred_array = _check_arrays(y_true, y_pred)
+    return float(np.mean((y_true_array - y_pred_array) ** 2))
 
 
 def rmse(y_true: np.ndarray, y_pred: np.ndarray) -> float:
@@ -196,6 +215,8 @@ def rsquare(y_true: np.ndarray, y_pred: np.ndarray) -> float:
     float
         Coefficient of determination.
     """
-    d = (np.asarray(y_true) - np.asarray(y_pred)) ** 2
-    D = (np.asarray(y_true) - np.asarray(y_true).mean()) ** 2
+    y_true_array, y_pred_array = _check_arrays(y_true, y_pred)
+
+    d = (y_true_array - y_pred_array) ** 2
+    D = (y_true_array - y_pred_array.mean()) ** 2
     return 1 - np.sum(d) / np.sum(D)
