@@ -43,6 +43,11 @@ def test_node_link(plus_node, minus_node):
         plus_node >> plus_node
 
 
+def test_empty_model_init():
+    model = Model()
+    assert model.is_empty
+
+
 def test_model_call(plus_node, minus_node):
 
     model = plus_node >> minus_node
@@ -178,6 +183,37 @@ def test_model_run(plus_node, minus_node):
             assert_array_equal(expected_minus[-1][np.newaxis, :], minus_node.state())
 
 
+def test_model_run_on_sequences(plus_node, minus_node):
+    input_node = Input()
+    branch1 = input_node >> plus_node
+    branch2 = input_node >> minus_node
+
+    model = branch1 & branch2
+
+    data = np.zeros((5, 3, 5))
+    res = model.run(data)
+
+    assert set(res.keys()) == {plus_node.name, minus_node.name}
+    assert len(res[plus_node.name]) == 5
+    assert len(res[minus_node.name]) == 5
+    assert res[plus_node.name][0].shape == (3, 5)
+
+    input_node = Input()
+    branch1 = input_node >> plus_node
+    branch2 = input_node >> minus_node
+
+    model = branch1 & branch2
+
+    data = [np.zeros((3, 5)), np.zeros((8, 5))]
+    res = model.run(data)
+
+    assert set(res.keys()) == {plus_node.name, minus_node.name}
+    assert len(res[plus_node.name]) == 2
+    assert len(res[minus_node.name]) == 2
+    assert res[plus_node.name][0].shape == (3, 5)
+    assert res[plus_node.name][1].shape == (8, 5)
+
+
 def test_model_feedback(plus_node, minus_node, feedback_node):
 
     model = plus_node >> feedback_node >> minus_node
@@ -318,36 +354,64 @@ def test_offline_fit_simple_model(offline_node, offline_node2, plus_node, minus_
 
 
 def test_offline_fit_simple_model_fb(
-    offline_node, offline_node2, plus_node, minus_node, feedback_node
+    basic_offline_node, offline_node2, plus_node, minus_node, feedback_node
 ):
 
-    model = plus_node >> feedback_node >> offline_node
-    feedback_node <<= offline_node
+    model = plus_node >> feedback_node >> basic_offline_node
+    feedback_node <<= basic_offline_node
 
     X = np.ones((5, 5)) * 0.5
     Y = np.ones((5, 5))
 
     model.fit(X, Y)
 
-    assert_array_equal(offline_node.b, np.array([7.5]))
+    assert_array_equal(basic_offline_node.b, np.array([9.3]))
+
+    model = plus_node >> feedback_node >> basic_offline_node
+    feedback_node <<= basic_offline_node
 
     X = np.ones((3, 5, 5)) * 0.5
     Y = np.ones((3, 5, 5))
 
     model.fit(X, Y)
 
-    assert_array_equal(offline_node.b, np.array([97.5]))
+    assert_array_equal(basic_offline_node.b, np.array([11.4]))
 
     model.fit(X, Y, reset=True)
 
-    assert_array_equal(offline_node.b, np.array([22.5]))
+    assert_array_equal(basic_offline_node.b, np.array([5.15]))
 
     res = model.run(X[0], reset=True)
 
-    exp = np.tile(np.array([26, 54.5, 85.5, 119, 155]), 5).reshape(5, 5).T
+    exp = np.tile(np.array([8.65, 19.8, 33.45, 49.6, 68.25]), 5).reshape(5, 5).T
 
     assert_array_equal(exp, res)
 
+    model = plus_node >> feedback_node >> basic_offline_node
+    feedback_node <<= basic_offline_node
+
+    X = np.ones((3, 5, 5)) * 0.5
+    Y = np.ones((3, 5, 5))
+
+    model.fit(X, Y, force_teachers=False)
+
+
+def test_offline_fit_complex(
+    basic_offline_node, offline_node2, plus_node, minus_node, feedback_node
+):
+
+    model = [plus_node >> basic_offline_node, plus_node] >> minus_node >> offline_node2
+
+    X = np.ones((5, 5, 5)) * 0.5
+    Y_1 = np.ones((5, 5, 5))
+    Y_2 = np.ones((5, 5, 10))  # after concat
+
+    model.fit(X, Y={"BasicOffline-0": Y_1, "Offline2-0": Y_2})
+
+    res = model.run(X[0])
+
+    assert res.shape == (5, 10)
+
 
 def test_online_train_simple(online_node, plus_node):
 
@@ -420,3 +484,44 @@ def test_online_train_teacher_nodes(online_node, plus_node, minus_node):
     model.train(X, minus_node, reset=True)
 
     assert_array_equal(online_node.b, np.array([108.0]))
+
+
+def test_model_return_states():
+    off = Offline(name="offline")
+    plus = PlusNode(name="plus")
+    minus = MinusNode(name="minus")
+    inverter = Inverter(name="inv")
+
+    model = plus >> [minus, off >> inverter]
+
+    X = np.ones((5, 5)) * 0.5
+    Y = np.ones((5, 5))
+
+    model.fit(X, Y)
+
+    res = model.run(X)
+
+    assert set(res.keys()) == {"minus", "inv"}
+
+    res = model.run(X, return_states="all")
+
+    assert set(res.keys()) == {"minus", "inv", "offline", "plus"}
+
+    res = model.run(X, return_states=["offline"])
+
+    assert set(res.keys()) == {"offline"}
+
+
+def test_multiinputs():
+    import numpy as np
+
+    from reservoirpy.nodes import Input, Reservoir
+
+    source1, source2 = Input(name="s1", input_dim=5,), Input(
+        name="s2",
+        input_dim=3,
+    )
+    res1, res2 = Reservoir(100), Reservoir(100)
+    # model = source1 >> [res1, res2] & source2 >> [res1, res2]
+    model = [source1, source2] >> res1 & [source1, source2] >> res2
+    outputs = model.run({"s1": np.ones((10, 5)), "s2": np.ones((10, 3))})
